import asyncio
import json
import os
import pickle
import shutil
from abc import ABC
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from hashlib import md5
from inspect import getsource
from pathlib import Path
from pickle import PicklingError
from tempfile import TemporaryDirectory
from typing import Any, TypeGuard, cast

from rich.console import Console

from .file_objects import BaseEntry, File

console = Console(
    log_path=False,
)


def _pickleable(obj):
    """
    Check if object is pickable
    """
    try:
        pickle.dumps(obj)
        return True
    except PicklingError:
        return False


def _is_valid_input(input: Any) -> TypeGuard[dict[str, Any] | None]:
    """
    Validate input

    key should be string
    value is BaseEntry, BaseTask or pickleable object

    or Map of these

    str
    int
    float
    BaseEntry
    BaseTask

    or Map / Sequence of these

    """
    if input is None:
        return True

    if not isinstance(input, dict):
        return False

    for key, val in input.items():
        if not isinstance(key, str):
            return False

        if not (
            BaseTask in val.__class__.__mro__
            or BaseEntry in val.__class__.__mro__
            or _pickleable(val)
        ):
            return False

    return True


def _is_valid_output(output: Any):
    """
    Validate output

    key should be string

    value should be File, Directory, str or pickleable object
    """
    if output is None:
        return True

    if not isinstance(output, dict):
        return False

    for key, val in output.items():
        if not isinstance(key, str):
            return False

        if not (BaseEntry in val.__class__.__mro__ or _pickleable(val)):
            return False

    return True


class BaseTask[INPUT, OUTPUT](ABC):
    _global_locks = defaultdict(asyncio.Lock)

    def __init__(
        self,
        cache_dir: str | Path,
        input: INPUT,
        worker: ThreadPoolExecutor | None = None,
        work_dir: str | Path | None = None,
        show_input: bool = False,
        show_output: bool = False,
    ):
        self._work_dir: Path | TemporaryDirectory | None = (
            Path(work_dir) if work_dir else None
        )
        self._cached_output: OUTPUT | None = None
        self._worker = worker
        self._cache_dir = Path(cache_dir)
        self._input = input
        self._show_input = show_input
        self._show_output = show_output

        self._hash: str | None = None

    async def task(self, input: INPUT) -> OUTPUT:
        _ = input
        raise NotImplementedError

    def result(self) -> OUTPUT:
        return asyncio.run(self())

    async def __call__(self) -> OUTPUT:
        async with BaseTask._global_locks[self.hash]:
            if self._cached_output is not None:
                return self._cached_output

            if self._check_cache():
                output = self._load_from_cache()
                console.log(
                    f"{self.__class__.__name__}: Cache found skipping",
                )
                if self._show_output:
                    console.log(
                        f"{self.__class__.__name__}: Input",
                        self._input,
                    )
                if self._show_output:
                    console.log(
                        f"{self.__class__.__name__}: Output",
                        output,
                    )
                return output

            await self._setup()

            prev_dir = os.getcwd()
            os.chdir(self.work_dir)

            if self._worker is None:
                console.log(f"{self.__class__.__name__}: Running with")
                if self._show_input:
                    console.log(
                        self._input,
                    )
                output = await self.task(self._input)
            else:

                def run_async_in_thread(async_func, *args, **kwargs):
                    console.log(f"{self.__class__.__name__}: Running with")
                    if self._show_input:
                        console.log(
                            self._input,
                        )
                    return asyncio.run(async_func(*args, **kwargs))

                loop = asyncio.get_event_loop()
                output = loop.run_in_executor(
                    self._worker, run_async_in_thread, self.task, self._input
                )
                output = await output

            if not _is_valid_output(output):
                raise ValueError(
                    "Output should be a dictionary with string keys and have pickleable values"
                )
            self._cache_output(output)
            self._cached_output = output

            os.chdir(prev_dir)

        if self._show_output:
            console.log(
                f"{self.__class__.__name__}: Output",
                output,
            )
        return output

        # return self.output

    @property
    async def output(self) -> OUTPUT:
        """
        return synchronous output
        """
        return await self()

    async def _setup(self):
        """Setup the Task

        This method will copy input files to work_dir
        and dependencies output files to work_dir
        """
        if not _is_valid_input(self._input):
            raise ValueError("Input should be a dictionary with string keys")

        if self._input is None:
            return

        for _, inval in self._input.items():
            if isinstance(inval, BaseEntry):
                inval._copy_to_dest(self.work_dir)

        dependent_tasks = []
        for _, inval in self._input.items():
            if isinstance(inval, BaseTask):
                dependent_tasks.append(inval())
        dependent_tasks_output = await asyncio.gather(*dependent_tasks)

        for output in dependent_tasks_output:
            for _, val in output.items():
                if isinstance(val, BaseEntry):
                    val.copy(self.work_dir)

    def _check_cache(self) -> bool:
        """
        Check if cache exists
        If corresponding hash directory or output files are not found return False
        """
        if os.path.exists(self._cache_dir / self.hash):
            return True
        return False

    def _load_from_cache(self) -> OUTPUT:
        """
        Load output from cache,

            File, Directory: set cache_path
            str            : read from file

        /key/data -> pickled data

        """
        output = {}

        if not os.path.exists(self._cache_dir / self.hash):
            raise FileNotFoundError(f"Cache for {self.hash} not found")

        for dir in os.listdir(self._cache_dir / self.hash):
            output[dir] = pickle.load(
                open(self._cache_dir / self.hash / dir / "data", "rb")
            )
        self._cached_output = cast(OUTPUT, output)
        return self._cached_output

    def _cache_output(self, output):
        if not isinstance(output, dict):
            return

        # remove if exists
        if os.path.exists(self._cache_dir / self.hash):
            shutil.rmtree(self._cache_dir / self.hash)

        os.makedirs(self._cache_dir / self.hash)
        for i, v in output.items():
            val: str | File = v
            cache_path: Path = self._cache_dir / self.hash / i
            if isinstance(val, BaseEntry):
                val._cache(cache_path)
            elif _pickleable(val):
                # save string to cache
                os.makedirs(cache_path, exist_ok=True)
                with open(cache_path / "data", "wb") as f:
                    f.write(pickle.dumps(val))

    @property
    def hash(self):
        """
        Calculate hash of the task
        Hash includes source code of the task and input text and file and directory content
        """
        return md5(self._dump_input().encode()).hexdigest()

    def _dump_input(self, indent: bool | int = False) -> str:
        """
        Dump input to as json string

        """

        if not _is_valid_input(self._input):
            raise ValueError(
                "Input should be a dictionary with string keys and have pickleable values"
            )

        ret_dict: dict[str, Any] = {
            "__lazypp_task_source__": md5(
                getsource(self.task.__code__).encode()
            ).hexdigest(),
        }

        if self._input is None:
            return json.dumps(ret_dict)

        for i, v in self._input.items():
            if BaseTask in v.__class__.__mro__:
                ret_dict[i] = v.hash
            elif BaseEntry in v.__class__.__mro__:
                ret_dict[i] = v._md5_hash().hexdigest()
            else:
                ret_dict[i] = md5(pickle.dumps(v)).hexdigest()

        return json.dumps(ret_dict, indent=indent)

    @property
    def work_dir(self):
        """
        This creates a temporary directory if work_dir is not set
        if work_dir is set the work_dir would not be deleted
        """
        if self._work_dir is None:
            self._work_dir = TemporaryDirectory()
        if isinstance(self._work_dir, TemporaryDirectory):
            return Path(self._work_dir.name)
        return self._work_dir



import os
import pickle
import shutil
from abc import ABC
from hashlib import md5
from pathlib import Path


def _is_outside_base(relative_path: Path) -> bool:
    """
    Check if relative path is outside base directory
    """
    depth = 0
    for part in relative_path.parts:
        if part == "..":
            depth -= 1
        elif part != ".":
            depth += 1
        if depth < 0:
            return True
    return False


class BaseEntry(ABC):
    def __init__(
        self,
        path: str | Path,
        *,
        copy: bool = False,
        dest: str | Path | None = None,
    ):
        self._copy = copy
        self._dest_path = Path(dest) if dest is not None else Path(path)
        self._src_path = Path(path)
        self._src_path = self._src_path.resolve()

        if not self._src_path.exists():
            raise FileNotFoundError(f"File not found: {self._src_path}")

        if _is_outside_base(self._dest_path):
            raise ValueError("File is outside base directory")

    @property
    def path(self):
        if self._dest_path:
            return self._dest_path
        return self._src_path

    def __str__(self):
        return str(self._src_path)

    def __repr__(self):
        return str(self._src_path)

    def _md5_hash(self):
        raise NotImplementedError

    def _copy_to_dest(self, work_dir: Path):
        _ = work_dir
        raise NotImplementedError

    def _cache(self, cache_dir: Path):
        _ = cache_dir
        raise NotImplementedError

    def copy(self, dest: Path):
        _ = dest
        raise NotImplementedError


class File(BaseEntry):
    def _md5_hash(self):
        ret = md5()
        with open(self._src_path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                ret.update(chunk)
        return ret

    def _copy_to_dest(self, work_dir: Path):
        if self._copy:
            self.copy(work_dir)

    def _cache(self, cache_dir: Path):
        """Cache file to cache directory"""
        dest = cache_dir / self._md5_hash().hexdigest()
        os.makedirs(dest.parent, exist_ok=True)

        shutil.copy(self._dest_path, cache_dir / self._md5_hash().hexdigest())
        self._src_path = dest

        # save self instance to cache directory
        with open(cache_dir / "data", "wb") as f:
            f.write(pickle.dumps(self))

    def copy(self, dest: Path):
        os.makedirs((dest / self.path).parent, exist_ok=True)
        shutil.copy(self._src_path, dest / self.path)


class Directory(BaseEntry):
    def _md5_hash(self):
        ret = md5()
        for root, _, files in os.walk(self._src_path):
            for file in files:
                file_path = os.path.join(root, file)
                with open(file_path, "rb") as f:
                    for chunk in iter(lambda: f.read(4096), b""):
                        ret.update(chunk)
        return ret

    def _copy_to_dest(self, work_dir: Path):
        if self._copy:
            self.copy(work_dir)

    def _cache(self, cache_dir: Path):
        """Cache directory to cache directory"""
        dest = cache_dir / self._md5_hash().hexdigest()
        os.makedirs(dest.parent, exist_ok=True)

        shutil.copytree(self._dest_path, dest)
        self._src_path = dest

        # save self instance to cache directory
        with open(cache_dir / "data", "wb") as f:
            f.write(pickle.dumps(self))

    def copy(self, dest: Path):
        os.makedirs((dest / self.path).parent, exist_ok=True)
        shutil.copytree(self._src_path, dest / self.path)


def load_from_cache(path: Path) -> BaseEntry:
    with open(path / "data", "rb") as f:
        obj = pickle.load(f)

    return obj



import os
from typing import TypedDict

import pytest

from lazypp import BaseTask, Directory, File
from lazypp.task import _is_valid_input, _is_valid_output

# def test_dump_task_input(tmpdir):
#     class TestTaskBase[INPUT, OUTPUT](BaseTask[INPUT, OUTPUT]):
#         def __init__(self, input: INPUT, output: OUTPUT):
#             super().__init__(
#                 cache_dir=tmpdir, work_dir=None, input=input, output=output
#             )
#
#     class TestTaskInput(TypedDict):
#         input1: int
#         input2: str
#         input3: list[float]
#
#     class Task(TestTaskBase[TestTaskInput, None]):
#         def task(self, input, output):
#             pass
#
#     task = Task(
#         input={"input1": 1, "input2": "2", "input3": [1.0, 2.0, 3.0]},
#         output=None,
#     )
#
#     assert task.hash == "49a7bff5a6b2c1769ba7d29e20c1c6a2"


@pytest.mark.parametrize(
    "valid, input",
    [
        (
            True,
            {
                "input1": 1,
                "input2": "2",
                "input3": [1.0, 2.0],
                "input4": ("a", "b"),
                "input5": {"a": 1, "b": 2},
            },
        ),
        (True, {"input": File("data/hello1.txt"), "input2": Directory("data/foo1")}),
        (
            True,
            {
                "input": [File("data/hello1.txt"), File("data/hello2.txt")],
                "input2": Directory("data/foo1"),
            },
        ),
        (False, {1: 1}),
        (False, {(1, 2): 1}),
    ],
)
def test_invalid_input(valid: bool, input: dict):
    assert _is_valid_input(input) == valid


# @pytest.mark.parametrize(
#     "valid, input",
#     [
#         (
#             True,
#             {
#                 "input1": 1,
#                 "input2": "2",
#                 "input3": [1.0, 2.0],
#                 "input4": ("a", "b"),
#                 "input5": {"a": 1, "b": 2},
#             },
#         ),
#         (True, {"input": File("test.txt"), "input2": Directory("foo")}),
#         (
#             True,
#             {
#                 "input": [File("test1.txt"), File("test2.txt")],
#                 "input2": Directory("foo"),
#             },
#         ),
#         (False, {1: 1}),
#         (False, {(1, 2): 1}),
#     ],
# )
# def test_invalid_output(valid: bool, input: dict):
#     assert _is_valid_output(input) == valid


def test_task(tmpdir):
    class TestInput(TypedDict):
        input1: int
        input2: str
        input3: list[float]

    class TestOutput(TypedDict):
        output1: File
        output2: Directory

    class TestTask(BaseTask[TestInput, TestOutput]):
        async def task(self, input) -> TestOutput:
            with open("output1.txt", "w") as f:
                f.write(str(input["input1"]))

            os.mkdir("output2")
            with open("output2/output2.txt", "w") as f:
                f.write(input["input2"])

            with open("output2/output3.txt", "w") as f:
                f.write(str(input["input3"]))

            return {"output1": File("output1.txt"), "output2": Directory("output2")}

    task = TestTask(cache_dir=tmpdir)

    print(task(input={"input1": 1, "input2": "2", "input3": [1.0, 2.0]}))



from pathlib import Path

import pytest

from lazypp import BaseEntry, Directory, File


def test_exceptions():
    file = BaseEntry("data")

    with pytest.raises(NotImplementedError):
        file._md5_hash()

    with pytest.raises(FileNotFoundError):
        File("data/hello3.txt")

    with pytest.raises(FileNotFoundError):
        Directory("data2")

    with pytest.raises(ValueError):
        File("../lazypp/__init__.py")

    with pytest.raises(ValueError):
        File("data/hello1.txt", dest="../dest_hello.txt")


def test_hash():
    file1 = File("data/hello1.txt")
    file2 = File("data/hello2.txt")

    assert file1._md5_hash().hexdigest() != file2._md5_hash().hexdigest()

    dir1 = Directory("data/foo1")
    dir2 = Directory("data/foo2")

    assert dir1._md5_hash().hexdigest() != dir2._md5_hash().hexdigest()


def test_dest():
    file = File("data/hello1.txt")
    assert file.path == Path("data/hello1.txt")

    file = File("data/hello1.txt", dest="dest_hello.txt")
    assert file.path == Path("dest_hello.txt")

    dir = Directory("data/foo1")
    assert dir.path == Path("data/foo1")

    dir = Directory("data/foo1", dest="dest_foo")
    assert dir.path == Path("dest_foo")



import asyncio
from pathlib import Path
from typing import TypedDict

from lazypp import BaseTask


class TestBaseTask[INPUT, OUTPUT](BaseTask[INPUT, OUTPUT]):
    def __init__(self, input: INPUT):
        super().__init__(
            cache_dir=Path("cache").resolve(),
            input=input,
        )


class Fin(TypedDict):
    your_name: str


class Fout(TypedDict):
    output: str


class Hello(TestBaseTask[Fin, Fout]):
    async def task(self, input: Fin) -> Fout:
        await asyncio.sleep(3)  # Some long running task
        return {"output": f"Hello, {input['your_name']}"}


ctask = Hello(
    input={"your_name": "John"},
)

print(ctask.result())
